from typing import Any, Dict, List, Optional, Text, Tuple, Type, Callable
from vi_nlp_core.ner.extractor import Extractor

from rasa.nlu.components import Component
from rasa.nlu.extractors.extractor import EntityExtractor
from rasa.nlu.config import RasaNLUModelConfig
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.nlu.training_data.message import Message


class PatternExtractor(EntityExtractor):
    name = "PatternExtractor"
    defaults = {
        "n_gram": 4,
        "load_dict": False,
        "mode": "pattern",
    }

    def __init__(
        self,
        component_config: Optional[Dict[Text, Any]] = None,
    ) -> None:
        super(PatternExtractor, self).__init__(component_config)

        self.extractor = Extractor(self.component_config["n_gram"], self.component_config["load_dict"])

    def train(
        self,
        training_data: TrainingData,
        config: Optional[RasaNLUModelConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Train this component.

        This is the components chance to train itself provided
        with the training data. The component can rely on
        any context attribute to be present, that gets created
        by a call to :meth:`components.Component.pipeline_init`
        of ANY component and
        on any context attributes created by a call to
        :meth:`components.Component.train`
        of components previous to this one."""
        pass

    def process(self, message: Message, **kwargs: Any) -> None:
        """Process an incoming message."""
        text = message.data.get('text')
        intent = message.data.get('intent')
        if text:
            output = self.extractor.extract_ner(text, intent['name'])
            old_entities = message.data.get("entities")
            for entity in output:
                entity['extractor'] = 'PatternExtractor'
                if entity['entity'] == 'time' and entity["value"]:
                    entity['value'] = f'{entity["value"][0]}:{entity["value"][1]}'
                    old_entities.append(entity)
                elif entity['entity'] == 'date_time' and entity["value"]:
                    first_date_extracted = entity["value"][0]
                    entity['value'] = f'{first_date_extracted[1]}/{first_date_extracted[2]}/{first_date_extracted[3]}'
                    old_entities.append(entity)
                else:
                    old_entities.append(entity)
            message.set("entities", old_entities, add_to_output=True)
        

    def persist(self, file_name: Text, model_dir: Text) -> Optional[Dict[Text, Any]]:
        """Persist this component to disk for future loading."""

        pass

    @classmethod
    def load(
        cls,
        meta: Dict[Text, Any],
        model_dir: Text,
        model_metadata: Optional["Metadata"] = None,
        cached_component: Optional["Component"] = None,
        **kwargs: Any,
    ) -> "Component":
        """Load this component from file."""

        if cached_component:
            return cached_component
        else:
            return cls(meta)



import datetime
import string
import re

from nltk import ngrams
from vi_nlp_core.utils.time_pattern import get_time_pattern
from vi_nlp_core.utils.preproces import Preprocess
from vietnam_number import w2n

class TimeMatcher:
    def __init__(self):
        self.default_hour = -1
        self.default_min = -1
        self.left_shift = 3
        self.right_shift = 4

        self.max_n_grams = 3
        self.absolute_pattern, self.am_pattern, self.pm_pattern = get_time_pattern()

        self.preprocessor = Preprocess()

    def extract_time(self, text):
        text = self.preprocess(text)
        # text = self.preprocessor.preprocess(text)
        raw_text = text
        text = text.replace('tiếng','giờ')

        time_result = self.extract_absolute_time(text) # không có 'giờ' pattern trong câu
        if time_result[2] != self.default_hour and time_result[3] != self.default_min:
            return self.output_format(time_result,'absolute_pattern')
        else:
            time_result = self.extract_relative_time(text, raw_text)        
        return self.output_format(time_result,'relative_pattern')

    def extract_relative_time(self, text, raw_text):
        hour = self.default_hour
        minute = self.default_min
        start = 0
        end = 0
        status = self.get_time_status(text)

        text = text.split(' ')
        raw_text = raw_text.split(' ')

        if 'giờ' not in text:
            # Maybe it contains 'phut'
            if 'phút' in text or 'phut' in text:
                start,end, hour, minute = self.no_hour_case(text)
                hour, minute = self.refine_hour_minute(hour,minute,status)
                # should be round time 
            return start,end, hour, minute

        hour_index = text.index('giờ')
        hour_range = text[max(0, hour_index - self.left_shift): hour_index]
        minute_range = text[hour_index + 1: min(hour_index + self.right_shift, len(text))]

        hour = self.get_hour(hour_range)
        minute = self.get_minute(minute_range)

        hour, minute = self.refine_hour_minute(hour,minute,status)
        # hour, minute = self.round_hour_minute_to_base(hour,minute)
        
        start = len(' '.join(raw_text[:hour_index-1])) # future fix mười một giờ - index
        if start != 0:
            start += 1
        if minute == 0:
            end = len(' '.join(raw_text[:hour_index + 1]))
        else:
            end = len(' '.join(raw_text[: hour_index + self.right_shift + 1]))
        return start,end, hour, minute

    def extract_absolute_time(self, text):
        status = self.get_time_status(text)
        hour = self.default_hour
        minute = self.default_min
        start = -1
        end = -1

        if 'giờ' in text.split(' '):
            return start, end, hour, minute

        absolute_pattern = self.absolute_pattern[0]
        times = [(x.group(), x.span()) for x in re.finditer(absolute_pattern, text)]
        if times:
            hours = []
            minutes = []
            starts = []
            ends = []
            for time in times:
                [hour, minute] = re.split('[:.hg]', time[0])
                if not re.search('\d', minute):
                    minute = 0
                (start, end) = time[1]
                hour, minute = self.refine_hour_minute(hour, minute, self.get_time_status(text[start: start+10]))
                # hour, minute = self.round_hour_minute_to_base(hour,minute)
                hours.append(hour)
                minutes.append(minute)
                starts.append(start)
                ends.append(end)   

        return start, end, hour, minute

    def get_time_status(self, text, minunte_range=None):
        '''
        Return:
            result - tuple (A,B,C)
                A : am(0) or pm(1)
                B : kém(-1) or hơn(1)
                C : có nữa/tiếp theo/kế tiếp/lát/lát nữa không ? (0/1)
        '''
        c_pattern = ['nữa','tiếp theo','kế tiếp','lát','lát nữa','kế','sắp tới','phút sau','phút nữa','giờ rưỡi sau','giờ sau']
        # min_pattern = ['sau']
        result = [0,0,0]
        status = 'am'

        for pattern in self.am_pattern:
            if text.find(pattern) != -1:
                result[0] = 0 #'am'
                break
        for pattern in self.pm_pattern:
            if text.find(pattern) != -1:
                result[0] = 1 #'pm'
                break

        if text.find('kém') != -1 or text.find('thiếu') != -1:
            result[1] = -1
        elif text.find('hơn') != -1: 
            result[1] = 1

        for pattern in c_pattern:
            if text.find(pattern) != -1:
                result[2] = 1 #yes
        
        return result

    def refine_hour_minute(self,hour,minute,status):
        try:
            hour = int(hour)
            minute = int(minute)
        except :
            return self.default_hour, self.default_min

        if hour != self.default_hour and minute == self.default_min:
            minute = 0
        # Check hour/minute is valid
        if hour == self.default_hour and minute == self.default_min:
            return hour, minute
        if status[0] == 1 and hour < 12:
            hour += 12
        
        if status[1] == -1:
            hour -= 1
            minute = 60 - int(minute)

        if status[2] == 1:
            now = datetime.datetime.now()
            hour = now.hour + int(hour)
            minute = now.minute + int(minute)
        
        if minute >= 60:
            minute -= 60
            hour += 1
        hour,minute = self.is_valid(hour,minute)
        return hour, minute

    def is_valid(self,hour,minute):
        if hour > 24 or hour < 0 or minute > 60 or minute < 0:
            return self.default_hour, self.default_min
        return hour, minute

    def get_hour(self,hour_range):
        # refine hour_range : padding punct
        hour_range = self.refine_hour_range(hour_range)

        if hour_range[-1].replace('h','').replace('g','').isdigit():
            return hour_range[-1].replace('h','').replace('g','')
        else:
            try:
                hour = w2n(' '.join(hour_range))
            except:
                hour = self.default_hour
            return hour

    def filter_minute_rage(self,minute_range):
        fil = [
            'ngày','phút','tháng','năm','buổi','thứ','chiều','sáng','trưa','tối','xế'
        ]
        for i in range(len(minute_range)):
            if minute_range[i] in fil:
                minute_range = minute_range[:i]
                return minute_range
        return minute_range
        
    def get_minute(self,minute_range):
        if minute_range == []:
            return 0
        minute_range = self.filter_minute_rage(minute_range)
        if minute_range == []:
            return 0
        elif minute_range[0].isdigit():
            return minute_range[0]
        elif minute_range[-1].isdigit():
            return minute_range[-1]
        elif 'rưỡi' in minute_range:
            return 30
        
        else:
            try:
                minute = w2n(' '.join(minute_range))
            except:
                minute = 0
            return minute
            
    def output_format(self,time_result,extractor):
        if time_result[2] == -1 or time_result[3] == -1:
            return {
                'entities':[]
            }
        else:
            return {
                'entities':[{
                    'start' : time_result[0],
                    'end' : time_result[1],
                    'entity' : 'time',
                    'value' : (time_result[2],time_result[3]),
                    'confidence' : 1.0,
                    'extractor' : extractor
                }]
            }

    def preprocess(self,text):
        text = text.replace('\n','').strip()
        text = re.sub('(?<! )(?=[,!?()])|(?<=[,!?()])(?! )', r' ', text)
        return text

    def no_hour_case(self,text): # list 
        hour = self.default_hour
        minute = self.default_min
        start = 0
        end = 0
        for i in range(len(text)):
            if text[i] == 'phút' or text[i] == 'phut':
                minute = self.get_minute(text[max(0,i - self.left_shift):i])
                start = len(" ".join(text[:i-1]).strip())
                if start != 0 :
                    start += 1
                end = len(" ".join(text[:i+1]).strip())
                hour = 0
                return start,end, hour, minute
        return start, end, hour, minute

    def refine_hour_range(self,hour_range):
        s = ' '.join(hour_range)
        s = s.translate(str.maketrans({key: " {0} ".format(key) for key in string.punctuation}))
        hour_range = s.split(' ')
        return hour_range

    def round_minute_to_base(self,minute,base):
        '''
            return: a float rounded to the nearest 15min interval
            if base is 15 for example
        '''
        if minute in [0,15,30,60]:
            return minute
        rounded = base * ((minute//base) + 1)
        
        return rounded

    def round_hour_minute_to_base(self,hour,minute,base=15):
        if hour == self.default_hour and minute == self.default_min:
            return hour, minute
        minute = self.round_minute_to_base(minute,base)
        if minute >= 60:
            minute -= 60
            hour += 1
        return hour, minute


if __name__ == '__main__':
    matcher = TimeMatcher()
    # text = '14:30 sáng thứ 7 tuần này'
    # print(matcher.extract_absolute_time(text))
    # text = '18:30 tiếng kém mười lăm'
    # print(matcher.extract_absolute_time(text))
    # text = 'ngày 7 tháng 6 lúc 14.15.00 giờ'
    # print(matcher.extract_absolute_time(text))
    
    text = 'tôi cần book lịch gấp vào 1 tiếng rưỡi sau'
    print(matcher.extract_relative_time(text))

    text = '4 tiếng sau họp tại trụ sở CA'
    print(matcher.extract_relative_time(text))


import re

from vi_nlp_core.utils.util import read_dict, get_ngram, tokenize
from vi_nlp_core.utils.pername_pattern import get_person_pattern
from vi_nlp_core.utils.preproces import Preprocess

class PernameMatcher:
    def __init__(self,n_gram = 4, dict_path='data/fullname.pkl', sw_path = 'data/stopwords.pkl',load_dict=False):

        self.max_n_gram = n_gram
        if load_dict:
            self.fullname_dict = read_dict(dict_path)
        else:
            self.fullname_dict = None
        
        self.stopwords_dict = read_dict(sw_path)

        self.person_explicit, self.person_pronoun, self.person_semi_pronoun, self.matches = self.build_person_name_pattern()

        self.preprocessor = Preprocess()
    
    def extract_person_name(self,utterance,mode='pattern',rt='relative'):
        '''
        1. Extract ngrams from (max_ngram,2)
        2. Check each ngram is in dictionary (aho-corasik)
        Input:
            utterance  - string
            mode - string
                dictionary : dictionary-based
                pattern : pattern matching
                hybrid : dictionary + pattern
            rt - string
                exact : 
                    yes --> NER
                    no  --> None
                relative:
                    yes --> NER
                    no  --> filter stopwords 
        Return:
            string      :   persone name
            'Invalid'   :   otherwise
        '''
        assert mode in ['dictionary','pattern','hybrid']
        # Normalize Uncicode
        # utterance = self.preprocessor.preprocess(utterance)
        utterance = tokenize(utterance).lower()
        if mode == 'dictionary':
            assert self.fullname_dict != None
            # Aho-corasik searching
            dict_result = self.extract_person_name_dict(utterance)
            return self.output_format(
                    output = dict_result, 
                    utterance = utterance,
                    extractor = mode,
                    entity = 'person_name',
                    rt= rt
                )
        elif mode == 'pattern':
            # Pattern matching
            pattern_result = self.extract_person_name_pattern(utterance)
            if pattern_result != 'Invalid':
                pattern_result = [pattern_result]
            return self.output_format(
                    output = pattern_result, 
                    utterance = utterance,
                    extractor = mode,
                    entity = 'person_name',
                    rt= rt
                )
        else:
            # Dictionay + pattern matching
            result = []
            dict_result = self.extract_person_name_dict(utterance)
            pattern_result = self.extract_person_name_pattern(utterance)
            if dict_result != 'Invalid':
                result.append(dict_result)
            if pattern_result != 'Invalid':
                result.append(pattern_result)
            result = list(set(result))
            if result != []:
                return  self.output_format(
                        output = result, 
                        utterance = utterance,
                        extractor = mode,
                        entity = 'person_name',
                        rt= rt
                    )
            else:
                return  self.output_format(
                        output = 'Invalid',
                        utterance = utterance,
                        extractor = mode,
                        entity = 'person_name',
                        rt= rt
                    )

    def extract_person_name_dict(self,utterance):
        '''
        1. Extract ngrams from (max_ngram,2)
        2. Check each ngram is in dictionary (aho-corasik)
        Input:
            utterance  - string
        Return:
            list      :   list of persone name
            'Invalid'   :   otherwise
        '''
        assert self.fullname_dict != None # Can not empty dictionary
        result = []
        for i in range(self.max_n_gram,2,-1):
            ngrams = get_ngram(utterance.lower(),i)
            for item in ngrams:
                item = ' '.join(item)
                if self.get_name(item) != 'not_exists' and item not in result:
                    result.append(item)
        if result != []:         
            return result   
        else:
            return "Invalid"
            
    def extract_person_name_pattern(self,utterance):
        '''
        Logic flow : 
            Pronoun --> semi pronoun --> explicit
        Input:
            utterance  - string
        Return:
            stirng      :   persone name
            'Invalid'   :   otherwise
        '''
        if any(x in utterance for x in self.matches):
            for pattern_list in self.person_pronoun:
                for pattern in pattern_list:
                    pronoun = re.search(pattern,utterance)
                    if pronoun != None and pronoun.group(1) != None:
                        return pronoun.group(1)
            semi_pronoun = self.person_semi_pronoun.search(utterance)
            if semi_pronoun != None and semi_pronoun.group(1) != None:
                return semi_pronoun.group(1)
            else:
                return 'Invalid' 
        else:
            explicit = self.person_explicit.search(utterance)
            try:
                if explicit != None and explicit.group(1) != None:
                    return explicit.group(1)
                else:
                    return 'Invalid'
            except:
                return 'Invalid'

    def get_name(self,s):
        try:
            return self.fullname_dict.get(s,'not_exists')
        except:
            print('Your dictionary is now empty')
            print('Ensure you have load the vocabulary for fullname_dict ! Otherwise you can run self.load_dict() update the dictionary .')
            return 'not_exists'

    def build_person_name_pattern(self):
        '''
        Input: 
            None
        Return:
            List of regex format pattern
        '''
        person_explicit, format_pronoun, person_semi_pronoun, matches = get_person_pattern()
        
        person_explicit = "|".join(person_explicit)
        person_semi_pronoun = "|".join(person_semi_pronoun)

        person_explicit = re.compile(person_explicit)
        person_semi_pronoun = re.compile(person_semi_pronoun)

        return person_explicit, format_pronoun, person_semi_pronoun, matches
    
    def output_format(self,output,utterance,extractor,entity,rt):
        if entity == 'person_name':
            if output == 'Invalid':
                if rt == 'exact':
                    return {
                        'entities' : []
                    }
                else: # Use stopwords for generate
                    ent = self.get_back_up_name(utterance)
                    entity = self.output_person_format(ent,utterance,extractor)
                    return {
                        'entities' : [entity]
                    }
            else:
                entities = []
                for ent in output:
                    entities.append(self.output_person_format(ent,utterance,extractor))
                return {
                    'entities' : entities
                }

    def output_person_format(self,output,utterance,extractor):
        '''
        1. Get start/end index
        2. Capitalize entity
        Input:
            output - string
                person name entity
            utterance - string
                raw input string from user
            extractor - string in ['dictionary','pattern','hybrid']
                method of extractor 
        '''
        start = utterance.find(output)
        end = start + len(output)

        output = output.split(' ')
        value = ' '.join([x.capitalize() for x in output])
        
        return {
                "start": start,
                "end": end,
                "entity": "person_name",
                "value": value,
                "confidence": 1.0,
                "extractor": extractor
                }

    def load_dict(self,dict_path='./vma_nlu/data/fullname.pkl'):
        print('Loading person name vocabulary ...')
        self.fullname_dict = read_dict(dict_path)
    
    def get_back_up_name(self,utterance):
        '''
        Get name in case pattern/dict-based is failed
        1. Get name based on popular first name
        2. Remove stopwords then return
        '''
        popular_first_name = ['nguyễn','trần','lê','phạm','hoàng','huỳnh','phan','vũ','võ','đặng','bùi','đỗ','hồ ','ngô','dương','lý']
        name = []
        name = []
        name_max_len = 5
        utterance = utterance.split(' ')
        # Get name based on popular first name
        for i, token in enumerate(utterance):
            if token in popular_first_name:
                name = utterance[i:min(i+5,len(utterance))] 
                break
        if name != []:
            # Post processing name : (TODO)
            return ' '.join(name)

        # Remove stopwords then return
        name = []
        for token in utterance:
            if self.get_stop_word(token) == 'not_exists':
                name.append(token)
        return ' '.join(name)

    def get_stop_word(self,s):
        return self.stopwords_dict.get(s,'not_exists')

    def preprocessing(self, text):
        return self.preprocessor.preprocess(text)


def get_time_pattern():
    # absolute 1 
    # 5:30 , 6:15 , 8:53 , 3:00 , 3.14.52 , 5.21

    # 11 am, 12pm, 1pm, 2:15pm, 3.30pm ,4.45, 5:30, 06:15, 07:30pm, 8:30 pm, 9.15 pm
    # abs_pattern =  [
    #     "(?=((?: |^|\s*)[0-2]?\d[:.hg]))",
    #     "(?=((?: |^|\s*)[0-2]?\d[:.hg ]?[0-5]\d?(?:|$)))",
    #     "(?=((?: |^|\s*)[0-2]?\d[:.hg ]?[0-5]\d?(?= giờ|tiếng)\d?(?:|$)))"
    # ]
    abs_pattern = [
        '((\d?\d[:.hg]\s*\d\d?)(?=\s|$|p|a))|((\d?\d[:.hg])(?=\s|$|p|a))'
    ]

    # contains giờ (logic)
    #   - sáng/chiều/tối (8 giờ chieu)
    #   - tám giờ chiều (số --> chữ)
    #   - 
    #   - 

    # extract AM or PM
    am_pattern = ['am','sáng', 'giờ sáng', 'buổi sáng', 'sunup', 'morning']
    pm_pattern = ['pm','tối','chiều','buổi chiều', 'buổi tối', 'giờ chiều', 'giờ tối' ,'trưa', 'buổi trưa', 'giờ trưa', 'xế chiều', 'chiều tà', 'noon','afternoon','evening']
    

    # special : cuoi buoi sang cuoi buoi chieu
    first_spec_pattern = {
        'đầu giờ sáng'      : '7h00',
        'đầu buổi sáng'     : '7h00',
        'đầu giờ chiều'     : '13h30', 
        'đầu buổi chiều'    : '13h30', 
        
    }
    mid_spec_pattern = {
        'giữa giờ sáng'     : '9h',
        'giữa buổi sáng'    : '9h',
        'giữa buổi chiều'   : '9h',
        'giữa giờ chiều'    : '9h',
    }

    last_spec_pattern = {
        'cuối giờ sáng'     : '11h00',
        'cuối buổi sáng'    : '11h00',
        'cuối giờ chiều'    : '17h00', 
        'cuối buổi chiều'   : '17h00',
    } 
    spec_pattern = [
        first_spec_pattern,
        mid_spec_pattern,
        last_spec_pattern
    ] 
    
    relative_pattern = []
    return abs_pattern, am_pattern, pm_pattern




import re
import json

from datetime import date
import datetime

from vi_nlp_core.utils.date_pattern import absolute, relative
DICT_PATH = "./vi_nlp_core/data/dictionary_normalize_date.json"

class DateMatcher(object):
    def __init__(self,dict_path) -> None:
        super().__init__()

        self.abs_pattern, self.wod, self.wod_vn, self.day_vn, self.month, self.month_vn, self.year, self.only_number, self.short_abs = absolute()
        
        self.abs_pattern = "|".join([x for x in self.abs_pattern])

        self.wod = "|".join([self.wod[0], self.wod_vn[0]])
        self.day = self.day_vn[0]
        self.month = "|".join([self.month[0], self.month_vn[0]])
        self.year = self.year[0]

        self.short_abs = self.short_abs[0]
        self.only_number = self.only_number[0]

        self.short_time, self.long_time, self.adj_pattern_top, self.adj_pattern_middle, self.adj_pattern_bot = relative()
        self.rel_pattern = "|".join([self.short_time[0], self.long_time[0]])

        self.adj_pattern_top = self.adj_pattern_top[0]
        self.adj_pattern_middle = self.adj_pattern_middle[0]
        self.adj_pattern_bot = self.adj_pattern_bot[0]

        self.week_days=["thứ 2","thứ 3","thứ 4","thứ 5","thứ 6","thứ 7","chủ nhật"]
        with open(dict_path, "r", encoding="utf-8") as f:
            self.dict_normalize = json.load(f)

    def extract_date(self, text):

        text = text.lower()

        value = []
        entities = []

        get_pattern_absolute = [(x.group(), x.span()) for x in re.finditer(self.abs_pattern, text)]
        get_pattern_relative = [(x.group(), x.span()) for x in re.finditer(self.rel_pattern, text)]

        wod = [(x.group(), x.span()) for x in re.finditer(self.wod, text)]
        day = [(x.group(), x.span()) for x in re.finditer(self.day, text)]
        # print(day)
        month = [(x.group(), x.span()) for x in re.finditer(self.month, text)]
        year = [(x.group(), x.span()) for x in re.finditer(self.year, text)]

        wod, day, month, year = self.normalize_date_2(wod, day, month, year)

        short_abs = [(x.group(), x.span()) for x in re.finditer(self.short_abs, text)]

        # Process
        
        # Absolute - Number
        if get_pattern_absolute:
            del short_abs
            short_abs = None
            val, ent = self.extract_date_abs_number(get_pattern_absolute, "long")
            value.extend(val)
            entities.extend(ent)
            return self.output_format(value=value, entities=entities, extractor="date_matcher")
        elif short_abs:
            val, ent = self.extract_date_abs_number(short_abs, "short")
            value.extend(val)
            entities.extend(ent)
            return self.output_format(value=value, entities=entities, extractor="date_matcher")
            

        if get_pattern_relative: # Relative
            val, ent  = self._map_relative_to_date(get_pattern_relative)
            # print(val)
            # print(ent)
            # Relative clear
            val, ent, flag = self.extract_date_rel_with_adj(val, ent, self.adj_pattern_top, self.adj_pattern_middle, self.adj_pattern_bot, text)
            # print(val)
            # print(ent)
            val, ent = self.extract_date_rel_clearly(val, ent, wod, day, flag)
            # print(val)
            # print(ent)
            value.extend(val)
            entities.extend(ent)
            return self.output_format(value=value, entities=entities, extractor="date_matcher")

        # Return None if not extract
        tmp = max(len(wod), len(day), len(month), len(year))
        if tmp == 0:
            return self.output_format(value=value, entities=entities, extractor="date_matcher")
        
        # Absolute - String and Number
        val, ent  = self.extract_date_abs_string_number(wod, day, month, year)
        value.extend(val)
        entities.extend(ent)

        return self.output_format(value=value, entities=entities, extractor="date_matcher")

    def _map_relative_to_date(self, get_pattern_relative):

        value = []
        entities = []
        
        for pattern, span in get_pattern_relative:
            if re.search("((qua)|(rồi))", pattern):
                
                if re.search("(ngày)|(ngay)|(ngayf)", pattern) or re.search("(hôm)|(hom)", pattern) or re.search("(sáng)|(sang)", pattern) or re.search("(trưa)|(trua)", pattern) or re.search("(chiều)|(chieu)", pattern) or re.search("(tối)|(toi)", pattern):
                    get_date = self._get_day(timedelta=1, mode="sub")
                    day = int(get_date[2])
                    month = int(get_date[1])
                    year = int(get_date[0])
                    wod = self.week_days[datetime.date(year, month, day).weekday()]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(wod, day, month, year)])
                    
                
                elif re.search("(tuần)|(tuan)", pattern):
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append(self._get_week(mode="sub"))

                elif re.search("(tháng)|(thang)", pattern):
                    get_month = self._get_month("sub")
                    day = get_month[2]
                    month = get_month[1]
                    year = get_month[0]
                    wod = [self.week_days[datetime.date(int(year[i]), int(month[i]), int(day[i])).weekday()] for i in range(len(day))]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(w, d, m, y) for w, d, m ,y in zip(wod, day, month, year)])
            
            elif re.search("(này)|(nay)", pattern):

                if re.search("(ngày)|(ngay)|(ngayf)", pattern) or re.search("(hôm)|(hom)", pattern) or re.search("(sáng)|(sang)", pattern) or re.search("(trưa)|(trua)", pattern) or re.search("(chiều)|(chieu)", pattern) or re.search("(tối)|(toi)", pattern):
                    get_date = self._get_day(timedelta=0, mode="add")
                    day = int(get_date[2])
                    month = int(get_date[1])
                    year = int(get_date[0])
                    wod = self.week_days[datetime.date(year, month, day).weekday()]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(wod, day, month, year)])

                elif re.search("(tuần)|(tuan)", pattern):
                    today = datetime.date.today()
                    weekday = today.weekday()
                    tmp = []
                    for i in range(7-int(weekday)):
                        get_date = self._get_day(timedelta=i, mode="add")
                        day = int(get_date[2])
                        month = int(get_date[1])
                        year = int(get_date[0])
                        wod = self.week_days[datetime.date(year, month, day).weekday()]
                        tmp.append((wod, day, month, year))
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append(tmp)

                elif re.search("(tháng)|(thang)", pattern):
                    get_month = self._get_month("current")
                    day = get_month[2]
                    month = get_month[1]
                    year = get_month[0]
                    wod = [self.week_days[datetime.date(int(year[i]), int(month[i]), int(day[i])).weekday()] for i in range(len(day))]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(w, d, m, y) for w, d, m ,y in zip(wod, day, month, year)])
            
            elif re.search("mai", pattern) or re.search("tới", pattern) or re.search("sau", pattern):
                
                if re.search("(ngày)|(ngay)|(ngayf)", pattern) or re.search("(hôm)|(hom)", pattern) or re.search("(sáng)|(sang)", pattern) or re.search("(trưa)|(trua)", pattern) or re.search("(chiều)|(chieu)", pattern) or re.search("(tối)|(toi)", pattern):
                    get_date = self._get_day(timedelta=1, mode="add")
                    day = int(get_date[2])
                    month = int(get_date[1])
                    year = int(get_date[0])
                    wod = self.week_days[datetime.date(year, month, day).weekday()]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(wod, day, month, year)])

                elif re.search("(tuần)|(tuan)", pattern):
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append(self._get_week(mode="add"))
                

                elif re.search("(tháng)|(thang)", pattern):
                    get_date = self._get_month("add")
                    day = get_date[2]
                    month = get_date[1]
                    year = get_date[0]
                    wod = [self.week_days[datetime.date(int(year[i]), int(month[i]), int(day[i])).weekday()] for i in range(len(day))]
                    entities.append({
                        "start": span[0],
                        "end": span[1]
                    })
                    value.append([(w, d, m, y) for w, d, m ,y in zip(wod, day, month, year)])

            elif re.search("mốt", pattern):
                get_date = self._get_day(timedelta=2, mode="add")
                day = int(get_date[2])
                month = int(get_date[1])
                year = int(get_date[0])
                wod = self.week_days[datetime.date(year, month, day).weekday()]
                entities.append({
                    "start": span[0],
                    "end": span[1]
                })
                value.append([(wod, day, month, year)])

            elif re.search("kia", pattern):
                get_date = self._get_day(timedelta=3, mode="add")
                day = int(get_date[2])
                month = int(get_date[1])
                year = int(get_date[0])
                wod = self.week_days[datetime.date(year, month, day).weekday()]
                entities.append({
                    "start": span[0],
                    "end": span[1]
                })
                value.append([(wod, day, month, year)])
            
        return value, entities


    def _get_month(self, mode):
        today = datetime.date.today()
        month = int(str(today).split("-")[1])
        year = int(str(today).split("-")[0])
        curr_day = int(str(today).split("-")[2])
        if mode == "add":
            month += 1
            if month in [1, 3, 5, 7, 8, 10, 12]:
                day = list(range(1, 32))
            elif month in [4, 6, 9, 11]:
                day = list(range(1,31))
            else: day = list(range(1, 29))
            month = [month]*len(day)
            year = [year]*len(day)
            return [year, month, day]
        elif mode == "sub":
            month -= 1
            if month in [1, 3, 5, 7, 8, 10, 12]:
                day = list(range(1, 32))
            elif month in [4, 6, 9, 11]:
                day = list(range(1,31))
            else: day = list(range(1, 29))
            month = [month]*len(day)
            year = [year]*len(day)
            return [year, month, day]
        elif mode == "current":
            if month in [1, 3, 5, 7, 8, 10, 12]:
                day = list(range(curr_day, 32))
            elif month in [4, 6, 9, 11]:
                day = list(range(curr_day, 31))
            else: day = list(range(curr_day, 29))
            month = [month]*len(day)
            year = [year]*len(day)
            return [year, month, day]

    def _get_day(self, today=datetime.date.today(), timedelta=1, mode = "sub"):
        if mode == "add":
            return str(today + datetime.timedelta(days=timedelta)).split("-")
        elif mode == "sub":
            return str(today - datetime.timedelta(days=timedelta)).split("-")
    
    def _get_week(self, mode):
        today = datetime.date.today()
        start_delta = datetime.timedelta(today.weekday())
        start_of_week = today - start_delta
        value = []
        for i in range(7,14):
            if mode == "add":
                get_date = self._get_day(start_of_week, timedelta=i, mode=mode)
                day = int(get_date[2])
                month = int(get_date[1])
                year = int(get_date[0])
                wod = self.week_days[datetime.date(year, month, day).weekday()]
                value.append((wod, day, month, year))
            elif mode == "sub":
                get_date = self._get_day(start_of_week, 14-i, mode=mode)
                day = int(get_date[2])
                month = int(get_date[1])
                year = int(get_date[0])
                wod = self.week_days[datetime.date(year, month, day).weekday()]
                value.append((wod, day, month, year))
        return value # [("thứ 2", 1,1,111), ("thứ 3", 2, 1, 1111) ... ("chủ nhật", 7, 1, 1111)]          

    def output_format(self, value, entities, extractor):
        if not value:
            return {
                "entities": []
            }
        else:
            result = {
                "entities": []
            }
            for i in range(len(value)):
                result["entities"].append(
                    {
                        "start": entities[i]["start"],
                        "end": entities[i]["end"],
                        "entity": "date_time",
                        "value": value[i],
                        "confidence": 1.0,
                        "extractor": extractor
                    }
                )
            return result

    def extract_date_abs_number(self, patterns, cat="long"): # [21/03/1997 21-03-1997 21_03-1997, ... ]
        value = []
        entities = []
        for pattern, span in patterns:
            tmp = re.split('(,|\s|\.|-|\/|_)', pattern.strip())
            day = int(tmp[0])
            month = int(tmp[2])
            flag = self.check_valid_date(day, month)
            if not flag:
                value.append([])
            else:
                if cat == "long":
                    year = str(tmp[4])
                    if len(year) == 4:
                        year = int(year)
                    elif len(year) == 2:
                        if int(year) > 50:
                            year = "19"+year
                        else:
                            year = "20"+year
                        year = int(year)
                    elif len(year) == 1:
                        year = int(datetime.datetime.now().year)
                    wod = self.week_days[datetime.date(year, month, day).weekday()]
                    value.append([self.normalize_date((wod, day, month, year))])
                else:
                    year = None
                    wod = None
                    value.append([self.normalize_date((wod, day, month, year))])
            entities.append({
                "start": span[0],
                "end": span[1]
            })
            
        return value, entities

    def normalize_date(self, input_date): # (WOD, DD, MM, YYYY)
        WOD = input_date[0]
        DD = int(input_date[1])
        MM = int(input_date[2])
        YYYY = input_date[3]
        return (WOD, DD, MM, YYYY)

    def extract_date_rel_clearly(self, val, ent, wod, day, flag): # Ex: thứ 5 tuần sau, thứ 2 tuần tới
        if not wod and not day:
            if flag:
                return val, ent
            else:
                values = []
                entities = []
                for vs, es in zip(val, ent):
                    v = vs[int(len(vs)/2)]
                    values.append([v])
                    entities.append(es)
                return values, entities
        entities = []
        values = []

        for v, e in zip(val, ent): # [[("thứ 2", 22, 2, 2222)], [("thứ 2", 22, 2, 2222), ("thứ 3", 22, 3, 2232)]] # [{"start": 10, "end": 20 },  {"start": 20, "end":25}]
            if len(v) == 1:
                entities.append(v)
                values.append(e)
                continue
            if wod:
                start_span = 100
                end_span = 0
                start_span = min([i[1][0] for i in wod])
                start_span = min(start_span, e['start'])
                end_span = max([i[1][1] for i in wod])
                end_span = max(end_span, e['end'])
                value = []
                tmp = [i[0] for i in v]
                tmp_1 = [i[0] for i in wod]
                for index, i in enumerate(tmp):
                    if i in tmp_1:
                        value.extend([v[index]])
                if not value:
                    value = v
                    entities.append(e)
                else:
                    entities.append({
                        "start": start_span,
                        "end": end_span
                    })
                    values.append(value)


            if day:
                value = []
                start_span = 100
                end_span = 0
                start_span = min([i[1][0] for i in day])
                start_span = min(start_span, e['start'])
                end_span = max([i[1][1] for i in day])
                end_span = max(end_span, e['end'])
                tmp = [i[1] for i in v]
                tmp_1 = [i[0] for i in day]
                for index, i in enumerate(tmp):
                    if i in tmp_1:
                        value.extend([v[index]])
                if not value:
                    value = v
                    entities.append(e)
                else:
                    entities.append({
                        "start": start_span,
                        "end": end_span
                    })
                    values.append(value)              
        return values, entities


    def extract_date_abs_string_number(self, wod, day, month, year):
        tmp = max(len(wod), len(day), len(month), len(year))
        # print(tmp)
        start_span = 100
        end_span = 0
        wod_tmp = [None] * tmp
        day_tmp = [None] * tmp
        month_tmp = [None] * tmp
        year_tmp = [None] * tmp
        if wod:
            wod = wod * tmp
            wod = wod[:tmp]
            wod_tmp = [i[0] for i in wod]
            start_span = min([j[1][0] if j[1][0] < start_span else start_span for j in wod])
            end_span = max([j[1][1] if j[1][1] > end_span else end_span for j in wod])
        if day:
            day = day * tmp
            day = day[:tmp]
            day_tmp = [i[0] for i in day]
            start_span = min([j[1][0] if j[1][0] < start_span else start_span for j in day])
            end_span = max([j[1][1] if j[1][1] > end_span else end_span for j in day])
        if month:
            month = month * tmp
            month = month[:tmp]
            month_tmp = [i[0] for i in month]
            start_span = min([j[1][0] if j[1][0] < start_span else start_span for j in month])
            end_span = max([j[1][1] if j[1][1] > end_span else end_span for j in month])
        if year:
            year = year * tmp
            year = year[:tmp]
            year_tmp = [i[0] for i in year]
            start_span = min([j[1][0] if j[1][0] < start_span else start_span for j in year])
            end_span = max([j[1][1] if j[1][1] > end_span else end_span for j in year])
        value = []
        entities = [{
            "start": start_span,
            "end": end_span
        }]
        value.append([(w, d, m, y) for w, d, m, y in zip(wod_tmp, day_tmp, month_tmp, year_tmp)])
        return value, entities

    def normalize_date_2(self, wod, day, month, year):
        if wod:
            wod_tmp = []
            for i in wod:
                value = re.sub('(thứ)|(thu)|(thuws)|(Thứ)', '', i[0])
                value = value.strip()
                value = self.dict_normalize.get(value, value)
                if value != "chủ nhật":
                    value = 'thứ '+ str(value)
                wod_tmp.append((value, i[1]))
            wod = wod_tmp
        if day:
            day_tmp = []
            for i in day:
                value = re.sub('(ngày)|(ngay)|(ngafy)', '', i[0])
                value = re.sub('(tháng)|(thang)', '', value)
                value = value.strip()
                value = self.dict_normalize.get(value, value)
                day_tmp.append((value, i[1]))
            day = day_tmp
        if month:
            month_tmp = []
            for i in month:
                value = re.sub('(tháng)|(thang)|(thasng)', '', i[0])
                value = value.strip()
                value = self.dict_normalize.get(value, value)
                month_tmp.append((value, i[1]))
            month = month_tmp
        if year:
            year_tmp = []
            for i in year:
                value = re.sub('(năm)|(nam)', '', i[0])
                value = value.strip()
                value = int(value)
                year_tmp.append((value, i[1]))
            year = year_tmp
        return wod, day, month, year

    def extract_date_rel_with_adj(self, val, ent, top, middle, bot, text):
        flag = False
        top = [(x.group(), x.span()) for x in re.finditer(top, text)]
        middle = [(x.group(), x.span()) for x in re.finditer(middle, text)]
        bot = [(x.group(), x.span()) for x in re.finditer(bot, text)]
        if not top and not middle and not bot:
            return val, ent, flag
        else:
            flag = True
            values = []
            entities = []
            for vs, es in zip(val, ent):
                if top:
                    for i in top:
                        v = vs[:round(len(vs)*0.25)]
                        # print(v)
                        values.append(v)
                        start_span = min(i[1][0], es["start"])
                        end_span = max(i[1][1], es["end"])
                        entities.append({
                            "start": start_span,
                            "end": end_span
                        })
                if middle:
                    for i in middle:
                        v = vs[round(len(vs)*0.25):round(len(vs)*0.75)]
                        values.append(v)
                        start_span = min(i[1][0], es["start"])
                        end_span = max(i[1][1], es["end"])
                        entities.append({
                            "start": start_span,
                            "end": end_span
                        })
                if bot:
                    for i in bot:
                        v = vs[round(len(vs)*0.75):]
                        values.append(v)
                        start_span = min(i[1][0], es["start"])
                        end_span = max(i[1][1], es["end"])
                        entities.append({
                            "start": start_span,
                            "end": end_span
                        })
            return values, entities, flag


    def check_valid_date(self, day, month):
        flag = False
        day_count_for_month = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        if month > 12 or month <=0 or int(day) > day_count_for_month[(month)]:
            return flag
        else: return True


import re

class Preprocess(object):
    def __init__(self) -> None:
        super().__init__()
        self.vowel = [['a', 'à', 'á', 'ả', 'ã', 'ạ', 'a'],
                ['ă', 'ằ', 'ắ', 'ẳ', 'ẵ', 'ặ', 'aw'],
                ['â', 'ầ', 'ấ', 'ẩ', 'ẫ', 'ậ', 'aa'],
                ['e', 'è', 'é', 'ẻ', 'ẽ', 'ẹ', 'e'],
                ['ê', 'ề', 'ế', 'ể', 'ễ', 'ệ', 'ee'],
                ['i', 'ì', 'í', 'ỉ', 'ĩ', 'ị', 'i'],
                ['o', 'ò', 'ó', 'ỏ', 'õ', 'ọ', 'o'],
                ['ô', 'ồ', 'ố', 'ổ', 'ỗ', 'ộ', 'oo'],
                ['ơ', 'ờ', 'ớ', 'ở', 'ỡ', 'ợ', 'ow'],
                ['u', 'ù', 'ú', 'ủ', 'ũ', 'ụ', 'u'],
                ['ư', 'ừ', 'ứ', 'ử', 'ữ', 'ự', 'uw'],
                ['y', 'ỳ', 'ý', 'ỷ', 'ỹ', 'ỵ', 'y']]

        self.vowel_to_idx = {}

        for i in range(len(self.vowel)):
            for j in range(len(self.vowel[i]) - 1):
                self.vowel_to_idx[self.vowel[i][j]] = (i, j)
    def is_valid_vietnam_word(self,word):
        chars = list(word)
        vowel_index = -1
        for index, char in enumerate(chars):
            x, y = self.vowel_to_idx.get(char, (-1, -1))
            if x != -1:
                if vowel_index == -1:
                    vowel_index = index
                else:
                    if index - vowel_index != 1:
                        return False
                    vowel_index = index
        return True
    
    def normalize_word_diacritic(self,word):
        """
        diacritic: á, à, ạ, ả, ã
        params:
            raw word
        return:
            word normalize
        """
        if not self.is_valid_vietnam_word(word):
            return word

        chars = list(word)
        diacritic = 0
        vowel_index = []
        qu_or_gi = False
        for index, char in enumerate(chars):
            x, y = self.vowel_to_idx.get(char, (-1, -1))
            if x == -1:
                continue
            elif x == 9:  # check qu
                if index != 0 and chars[index - 1] == 'q':
                    chars[index] = 'u'
                    qu_or_gi = True
            elif x == 5:  # check gi
                if index != 0 and chars[index - 1] == 'g':
                    chars[index] = 'i'
                    qu_or_gi = True
            if y != 0:
                diacritic = y
                chars[index] = self.vowel[x][0]
            if not qu_or_gi or index != 1:
                vowel_index.append(index)
        if len(vowel_index) < 2:
            if qu_or_gi:
                if len(chars) == 2:
                    x, y = self.vowel_to_idx.get(chars[1])
                    chars[1] = self.vowel[x][diacritic]
                else:
                    x, y = self.vowel_to_idx.get(chars[2], (-1, -1))
                    if x != -1:
                        chars[2] = self.vowel[x][diacritic]
                    else:
                        chars[1] = self.vowel[5][diacritic] if chars[1] == 'i' else self.vowel[9][diacritic]
                return ''.join(chars)
            return word
        
        for index in vowel_index:
            x, y = self.vowel_to_idx[chars[index]]
            if x == 4 or x == 8:  # ê, ơ
                chars[index] = self.vowel[x][diacritic]
                # for index2 in vowel_index:
                #     if index2 != index:
                #         x, y = vowel_to_idx[chars[index]]
                #         chars[index2] = vowel[x][0]
                return ''.join(chars)

        if len(vowel_index) == 2:
            if vowel_index[-1] == len(chars) - 1:
                x, y = self.vowel_to_idx[chars[vowel_index[0]]]
                chars[vowel_index[0]] = self.vowel[x][diacritic]
                # x, y = vowel_to_idx[chars[vowel_index[1]]]
                # chars[vowel_index[1]] = vowel[x][0]
            else:
                # x, y = vowel_to_idx[chars[vowel_index[0]]]
                # chars[vowel_index[0]] = vowel[x][0]
                x, y = self.vowel_to_idx[chars[vowel_index[1]]]
                chars[vowel_index[1]] = self.vowel[x][diacritic]
        else:
            # x, y = vowel_to_idx[chars[vowel_index[0]]]
            # chars[vowel_index[0]] = vowel[x][0]
            x, y = self.vowel_to_idx[chars[vowel_index[1]]]
            chars[vowel_index[1]] = self.vowel[x][diacritic]
            # x, y = vowel_to_idx[chars[vowel_index[2]]]
            # chars[vowel_index[2]] = vowel[x][0]
        return ''.join(chars)

    def normalize_diacritic(self,text):
        """
        normalize diacritic
        params:
            crawl text
        return:
            text normalize
        """
        sentence = text.lower()
        words = sentence.split()
        for index, word in enumerate(words):
            cw = re.sub(r'(^\W*)([\w.]*\w+)(\W*$)', r'\1/\2/\3', word).split('/')
            if len(cw) == 3:
                cw[1] = self.normalize_word_diacritic(cw[1])
            words[index] = ''.join(cw)
        return ' '.join(words)

    def loaddicchar(self):
        dic = {}
        char1252 = 'à|á|ả|ã|ạ|ầ|ấ|ẩ|ẫ|ậ|ằ|ắ|ẳ|ẵ|ặ|è|é|ẻ|ẽ|ẹ|ề|ế|ể|ễ|ệ|ì|í|ỉ|ĩ|ị|ò|ó|ỏ|õ|ọ|ồ|ố|ổ|ỗ|ộ|ờ|ớ|ở|ỡ|ợ|ù|ú|ủ|ũ|ụ|ừ|ứ|ử|ữ|ự|ỳ|ý|ỷ|ỹ|ỵ|À|Á|Ả|Ã|Ạ|Ầ|Ấ|Ẩ|Ẫ|Ậ|Ằ|Ắ|Ẳ|Ẵ|Ặ|È|É|Ẻ|Ẽ|Ẹ|Ề|Ế|Ể|Ễ|Ệ|Ì|Í|Ỉ|Ĩ|Ị|Ò|Ó|Ỏ|Õ|Ọ|Ồ|Ố|Ổ|Ỗ|Ộ|Ờ|Ớ|Ở|Ỡ|Ợ|Ù|Ú|Ủ|Ũ|Ụ|Ừ|Ứ|Ử|Ữ|Ự|Ỳ|Ý|Ỷ|Ỹ|Ỵ'.split(
            '|')
        charutf8 = "à|á|ả|ã|ạ|ầ|ấ|ẩ|ẫ|ậ|ằ|ắ|ẳ|ẵ|ặ|è|é|ẻ|ẽ|ẹ|ề|ế|ể|ễ|ệ|ì|í|ỉ|ĩ|ị|ò|ó|ỏ|õ|ọ|ồ|ố|ổ|ỗ|ộ|ờ|ớ|ở|ỡ|ợ|ù|ú|ủ|ũ|ụ|ừ|ứ|ử|ữ|ự|ỳ|ý|ỷ|ỹ|ỵ|À|Á|Ả|Ã|Ạ|Ầ|Ấ|Ẩ|Ẫ|Ậ|Ằ|Ắ|Ẳ|Ẵ|Ặ|È|É|Ẻ|Ẽ|Ẹ|Ề|Ế|Ể|Ễ|Ệ|Ì|Í|Ỉ|Ĩ|Ị|Ò|Ó|Ỏ|Õ|Ọ|Ồ|Ố|Ổ|Ỗ|Ộ|Ờ|Ớ|Ở|Ỡ|Ợ|Ù|Ú|Ủ|Ũ|Ụ|Ừ|Ứ|Ử|Ữ|Ự|Ỳ|Ý|Ỷ|Ỹ|Ỵ".split(
            '|')
        for i in range(len(char1252)):
            dic[char1252[i]] = charutf8[i]
        return dic

    def covert_unicode(self, txt):
        dicchar = self.loaddicchar()
        return re.sub(
            r'à|á|ả|ã|ạ|ầ|ấ|ẩ|ẫ|ậ|ằ|ắ|ẳ|ẵ|ặ|è|é|ẻ|ẽ|ẹ|ề|ế|ể|ễ|ệ|ì|í|ỉ|ĩ|ị|ò|ó|ỏ|õ|ọ|ồ|ố|ổ|ỗ|ộ|ờ|ớ|ở|ỡ|ợ|ù|ú|ủ|ũ|ụ|ừ|ứ|ử|ữ|ự|ỳ|ý|ỷ|ỹ|ỵ|À|Á|Ả|Ã|Ạ|Ầ|Ấ|Ẩ|Ẫ|Ậ|Ằ|Ắ|Ẳ|Ẵ|Ặ|È|É|Ẻ|Ẽ|Ẹ|Ề|Ế|Ể|Ễ|Ệ|Ì|Í|Ỉ|Ĩ|Ị|Ò|Ó|Ỏ|Õ|Ọ|Ồ|Ố|Ổ|Ỗ|Ộ|Ờ|Ớ|Ở|Ỡ|Ợ|Ù|Ú|Ủ|Ũ|Ụ|Ừ|Ứ|Ử|Ữ|Ự|Ỳ|Ý|Ỷ|Ỹ|Ỵ',
            lambda x: dicchar[x.group()], txt)

    def preprocess(self, text):
        """
        pipeline normalize encoding + normalize diacritic
        params:
            raw text
        return:
            normalization text
        """
        norm_enc = self.covert_unicode(text)
        return self.normalize_diacritic(norm_enc)
        